@@ -32,29 +32,42 @@ import threading
 import uuid
 import warnings
 
+import debtcollector
 from positional import positional
 
-
 _request_store = threading.local()
 
 # These arguments will be passed to a new context from the first available
 # header to support backwards compatibility.
-_ENVIRON_HEADERS = {'auth_token': ['HTTP_X_AUTH_TOKEN',
-                                   'HTTP_X_STORAGE_TOKEN'],
-                    'user': ['HTTP_X_USER_ID',
-                             'HTTP_X_USER'],
-                    'tenant': ['HTTP_X_PROJECT_ID',
-                               'HTTP_X_TENANT_ID',
-                               'HTTP_X_TENANT'],
-                    'user_domain': ['HTTP_X_USER_DOMAIN_ID'],
-                    'project_domain': ['HTTP_X_PROJECT_DOMAIN_ID'],
-                    'user_name': ['HTTP_X_USER_NAME'],
-                    'project_name': ['HTTP_X_PROJECT_NAME',
-                                     'HTTP_X_TENANT_NAME'],
-                    'user_domain_name': ['HTTP_X_USER_DOMAIN_NAME'],
-                    'project_domain_name': ['HTTP_X_PROJECT_DOMAIN_NAME'],
-                    'request_id': ['openstack.request_id'],
-                    }
+_ENVIRON_HEADERS = {
+    'auth_token': ['HTTP_X_AUTH_TOKEN',
+                   'HTTP_X_STORAGE_TOKEN'],
+    'user': ['HTTP_X_USER_ID',
+             'HTTP_X_USER'],
+    'tenant': ['HTTP_X_PROJECT_ID',
+               'HTTP_X_TENANT_ID',
+               'HTTP_X_TENANT'],
+    'user_domain': ['HTTP_X_USER_DOMAIN_ID'],
+    'project_domain': ['HTTP_X_PROJECT_DOMAIN_ID'],
+    'user_name': ['HTTP_X_USER_NAME'],
+    'project_name': ['HTTP_X_PROJECT_NAME',
+                     'HTTP_X_TENANT_NAME'],
+    'user_domain_name': ['HTTP_X_USER_DOMAIN_NAME'],
+    'project_domain_name': ['HTTP_X_PROJECT_DOMAIN_NAME'],
+    'request_id': ['openstack.request_id'],
+    'global_request_id': ['openstack.global_request_id'],
+
+
+    'service_token': ['HTTP_X_SERVICE_TOKEN'],
+    'service_user_id': ['HTTP_X_SERVICE_USER_ID'],
+    'service_user_name': ['HTTP_X_SERVICE_USER_NAME'],
+    'service_user_domain_id': ['HTTP_X_SERVICE_USER_DOMAIN_ID'],
+    'service_user_domain_name': ['HTTP_X_SERVICE_USER_DOMAIN_NAME'],
+    'service_project_id': ['HTTP_X_SERVICE_PROJECT_ID'],
+    'service_project_name': ['HTTP_X_SERVICE_PROJECT_NAME'],
+    'service_project_domain_id': ['HTTP_X_SERVICE_PROJECT_DOMAIN_ID'],
+    'service_project_domain_name': ['HTTP_X_SERVICE_PROJECT_DOMAIN_NAME'],
+}
 
 
 def generate_request_id():
@@ -118,6 +131,44 @@ class _DeprecatedPolicyValues(collections.MutableMapping):
         return d
 
 
+# FIXME(dims): Skip deprecation warnings until Pike (Version 2.14)
+_log_deprecation_warnings = False
+
+
+def _moved_msg(new_name, old_name):
+    if old_name:
+        deprecated_msg = "Property '%(old_name)s' has moved to '%(new_name)s'"
+        deprecated_msg = deprecated_msg % {'old_name': old_name,
+                                           'new_name': new_name}
+
+        # FIXME(dims): Skip deprecation warnings until Pike (Version 2.14)
+        if _log_deprecation_warnings:
+            debtcollector.deprecate(deprecated_msg,
+                                    version='2.6',
+                                    removal_version='3.0',
+                                    stacklevel=5)
+
+
+def _moved_property(new_name, old_name=None, target=None):
+
+    if not target:
+        target = new_name
+
+    def getter(self):
+        _moved_msg(new_name, old_name)
+        return getattr(self, target)
+
+    def setter(self, value):
+        _moved_msg(new_name, old_name)
+        setattr(self, target, value)
+
+    def deleter(self):
+        _moved_msg(new_name, old_name)
+        delattr(self, target)
+
+    return property(getter, setter, deleter)
+
+
 class RequestContext(object):
 
     """Helper class to represent useful information about a request context.
@@ -129,13 +180,37 @@ class RequestContext(object):
     user_idt_format = u'{user} {tenant} {domain} {user_domain} {p_domain}'
 
     @positional(enforcement=positional.WARN)
-    def __init__(self, auth_token=None, user=None, tenant=None, domain=None,
-                 user_domain=None, project_domain=None, is_admin=False,
-                 read_only=False, show_deleted=False, request_id=None,
-                 resource_uuid=None, overwrite=True, roles=None,
-                 user_name=None, project_name=None, domain_name=None,
-                 user_domain_name=None, project_domain_name=None,
-                 is_admin_project=True):
+    def __init__(self,
+                 auth_token=None,
+                 user=None,
+                 tenant=None,
+                 domain=None,
+                 user_domain=None,
+                 project_domain=None,
+                 is_admin=False,
+                 read_only=False,
+                 show_deleted=False,
+                 request_id=None,
+                 resource_uuid=None,
+                 overwrite=True,
+                 roles=None,
+                 user_name=None,
+                 project_name=None,
+                 domain_name=None,
+                 user_domain_name=None,
+                 project_domain_name=None,
+                 is_admin_project=True,
+                 service_token=None,
+                 service_user_id=None,
+                 service_user_name=None,
+                 service_user_domain_id=None,
+                 service_user_domain_name=None,
+                 service_project_id=None,
+                 service_project_name=None,
+                 service_project_domain_id=None,
+                 service_project_domain_name=None,
+                 service_roles=None,
+                 global_request_id=None):
         """Initialize the RequestContext
 
         :param overwrite: Set to False to ensure that the greenthread local
@@ -145,18 +220,18 @@ class RequestContext(object):
                                  True for backwards compatibility.
         :type is_admin_project: bool
         """
+        # setting to private variables to avoid triggering subclass properties
+        self._user_id = user
+        self._project_id = tenant
+        self._domain_id = domain
+        self._user_domain_id = user_domain
+        self._project_domain_id = project_domain
+
         self.auth_token = auth_token
-        self.user = user
         self.user_name = user_name
-        # NOTE (rbradfor):  tenant will become project
-        # See spec discussion on https://review.openstack.org/#/c/290907/
-        self.tenant = tenant
         self.project_name = project_name
-        self.domain = domain
         self.domain_name = domain_name
-        self.user_domain = user_domain
         self.user_domain_name = user_domain_name
-        self.project_domain = project_domain
         self.project_domain_name = project_domain_name
         self.is_admin = is_admin
         self.is_admin_project = is_admin_project
@@ -164,12 +239,44 @@ class RequestContext(object):
         self.show_deleted = show_deleted
         self.resource_uuid = resource_uuid
         self.roles = roles or []
+
+        self.service_token = service_token
+        self.service_user_id = service_user_id
+        self.service_user_name = service_user_name
+        self.service_user_domain_id = service_user_domain_id
+        self.service_user_domain_name = service_user_domain_name
+        self.service_project_id = service_project_id
+        self.service_project_name = service_project_name
+        self.service_project_domain_id = service_project_domain_id
+        self.service_project_domain_name = service_project_domain_name
+        self.service_roles = service_roles or []
+
         if not request_id:
             request_id = generate_request_id()
         self.request_id = request_id
+        self.global_request_id = global_request_id
         if overwrite or not get_current():
             self.update_store()
 
+    # NOTE(jamielennox): To prevent circular lookups on subclasses that might
+    # point user to user_id we make user/user_id tenant/project_id etc point
+    # to the same private variable rather than each other.
+    tenant = _moved_property('project_id', 'tenant', target='_project_id')
+    user = _moved_property('user_id', 'user', target='_user_id')
+    domain = _moved_property('domain_id', 'domain', target='_domain_id')
+    user_domain = _moved_property('user_domain_id',
+                                  'user_domain',
+                                  target='_user_domain_id')
+    project_domain = _moved_property('project_domain_id',
+                                     'project_domain',
+                                     target='_project_domain_id')
+
+    user_id = _moved_property('_user_id')
+    project_id = _moved_property('_project_id')
+    domain_id = _moved_property('_domain_id')
+    user_domain_id = _moved_property('_user_domain_id')
+    project_domain_id = _moved_property('_project_domain_id')
+
     def update_store(self):
         """Store the context in the current thread."""
         _request_store.context = self
@@ -191,32 +298,38 @@ class RequestContext(object):
         # of our standard ones. This object acts like a dict but only values
         # from oslo.policy don't show a warning.
         return _DeprecatedPolicyValues({
-            'user_id': self.user,
-            'user_domain_id': self.user_domain,
-            'project_id': self.tenant,
-            'project_domain_id': self.project_domain,
+            'user_id': self.user_id,
+            'user_domain_id': self.user_domain_id,
+            'project_id': self.project_id,
+            'project_domain_id': self.project_domain_id,
             'roles': self.roles,
-            'is_admin_project': self.is_admin_project})
+            'is_admin_project': self.is_admin_project,
+            'service_user_id': self.service_user_id,
+            'service_user_domain_id': self.service_user_domain_id,
+            'service_project_id': self.service_project_id,
+            'service_project_domain_id': self.service_project_domain_id,
+            'service_roles': self.service_roles})
 
     def to_dict(self):
         """Return a dictionary of context attributes."""
-        user_idt = (
-            self.user_idt_format.format(user=self.user or '-',
-                                        tenant=self.tenant or '-',
-                                        domain=self.domain or '-',
-                                        user_domain=self.user_domain or '-',
-                                        p_domain=self.project_domain or '-'))
+        user_idt = self.user_idt_format.format(
+            user=self.user_id or '-',
+            tenant=self.project_id or '-',
+            domain=self.domain_id or '-',
+            user_domain=self.user_domain_id or '-',
+            p_domain=self.project_domain_id or '-')
 
-        return {'user': self.user,
-                'tenant': self.tenant,
-                'domain': self.domain,
-                'user_domain': self.user_domain,
-                'project_domain': self.project_domain,
+        return {'user': self.user_id,
+                'tenant': self.project_id,
+                'domain': self.domain_id,
+                'user_domain': self.user_domain_id,
+                'project_domain': self.project_domain_id,
                 'is_admin': self.is_admin,
                 'read_only': self.read_only,
                 'show_deleted': self.show_deleted,
                 'auth_token': self.auth_token,
                 'request_id': self.request_id,
+                'global_request_id': self.global_request_id,
                 'resource_uuid': self.resource_uuid,
                 'roles': self.roles,
                 'user_identity': user_idt,
@@ -232,6 +345,16 @@ class RequestContext(object):
         values.update(self.to_dict())
         return values
 
+    @property
+    def global_id(self):
+        """Return a sensible value for global_id to pass on.
+
+        When we want to make a call with to another service, it's
+        important that we try to use global_request_id if available,
+        and fall back to the locally generated request_id if not.
+        """
+        return self.global_request_id or self.request_id
+
     @classmethod
     def from_dict(cls, values, **kwargs):
         """Construct a context object from a provided dictionary."""
@@ -245,6 +368,7 @@ class RequestContext(object):
         kwargs.setdefault('read_only', values.get('read_only', False))
         kwargs.setdefault('show_deleted', values.get('show_deleted', False))
         kwargs.setdefault('request_id', values.get('request_id'))
+        kwargs.setdefault('global_request_id', values.get('global_request_id'))
         kwargs.setdefault('resource_uuid', values.get('resource_uuid'))
         kwargs.setdefault('roles', values.get('roles'))
         kwargs.setdefault('user_name', values.get('user_name'))
@@ -286,6 +410,11 @@ class RequestContext(object):
             roles = [r.strip() for r in roles.split(',')] if roles else []
             kwargs['roles'] = roles
 
+        if 'service_roles' not in kwargs:
+            roles = environ.get('HTTP_X_SERVICE_ROLES')
+            roles = [r.strip() for r in roles.split(',')] if roles else []
+            kwargs['service_roles'] = roles
+
         if 'is_admin_project' not in kwargs:
             # NOTE(jamielennox): we default is_admin_project to true because if
             # nothing is provided we have to assume it is the admin project to
